from enum import Enum


class SQuADTask(Enum):
    PHARMACY = "Pharmacy"
    NORMANS = "Normans"
    HUGUENOT = "Huguenot"
    DOCTOR_WHO = "Doctor_Who"
    OIL_CRISIS_1973 = "1973_oil_crisis"
    COMPUTATIONAL_COMPLEXITY_THEORY = "Computational_complexity_theory"
    WARSAW = "Warsaw"
    AMERICAN_BROADCASTING_COMPANY = "American_Broadcasting_Company"
    CHLOROPLAST = "Chloroplast"
    APOLLO_PROGRAM = "Apollo_program"
    TEACHER = "Teacher"
    MARTIN_LUTHER = "Martin_Luther"
    ECONOMIC_INEQUALITY = "Economic_inequality"
    YUAN_DYNASTY = "Yuan_dynasty"
    SCOTTISH_PARLIAMENT = "Scottish_Parliament"
    ISLAMISM = "Islamism"
    UNITED_METHODIST_CHURCH = "United_Methodist_Church"
    IMMUNE_SYSTEM = "Immune_system"
    NEWCASTLE_UPON_TYNE = "Newcastle_upon_Tyne"
    CTENOPHORA = "Ctenophora"
    FRESNO_CALIFORNIA = "Fresno,_California"
    STEAM_ENGINE = "Steam_engine"
    PACKET_SWITCHING = "Packet_switching"
    FORCE = "Force"
    JACKSONVILLE_FLORIDA = "Jacksonville,_Florida"
    EUROPEAN_UNION_LAW = "European_Union_law"
    SUPER_BOWL_50 = "Super_Bowl_50"
    VICTORIA_AND_ALBERT_MUSEUM = "Victoria_and_Albert_Museum"
    BLACK_DEATH = "Black_Death"
    CONSTRUCTION = "Construction"
    SKY_UK = "Sky_(United_Kingdom)"
    UNIVERSITY_OF_CHICAGO = "University_of_Chicago"
    VICTORIA_AUSTRALIA = "Victoria_(Australia)"
    FRENCH_AND_INDIAN_WAR = "French_and_Indian_War"
    IMPERIALISM = "Imperialism"
    PRIVATE_SCHOOL = "Private_school"
    GEOLOGY = "Geology"
    HARVARD_UNIVERSITY = "Harvard_University"
    RHINE = "Rhine"
    PRIME_NUMBER = "Prime_number"
    INTERGOVERNMENTAL_PANEL_ON_CLIMATE_CHANGE = (
        "Intergovernmental_Panel_on_Climate_Change"
    )
    AMAZON_RAINFOREST = "Amazon_rainforest"
    KENYA = "Kenya"
    SOUTHERN_CALIFORNIA = "Southern_California"
    NIKOLA_TESLA = "Nikola_Tesla"
    CIVIL_DISOBEDIENCE = "Civil_disobedience"
    GENGHIS_KHAN = "Genghis_Khan"
    OXYGEN = "Oxygen"
