# Internal GEPA constants - not exposed to users
GEPA_MIN_DELTA: float = 0.0
GEPA_TIE_TOLERANCE: float = 1e-9
GEPA_REWRITE_INSTRUCTION_MAX_CHARS: int = 4096

# Internal MIPROV2 constants - not exposed to users
MIPROV2_MIN_DELTA: float = 0.0
MIPROV2_REWRITE_INSTRUCTION_MAX_CHARS: int = 4096
MIPROV2_DEFAULT_NUM_CANDIDATES: int = 10
MIPROV2_DEFAULT_NUM_TRIALS: int = 20
MIPROV2_DEFAULT_MINIBATCH_SIZE: int = 25
MIPROV2_DEFAULT_MINIBATCH_FULL_EVAL_STEPS: int = 10
MIPROV2_DEFAULT_MAX_BOOTSTRAPPED_DEMOS: int = 4
MIPROV2_DEFAULT_MAX_LABELED_DEMOS: int = 4
MIPROV2_DEFAULT_NUM_DEMO_SETS: int = 5

# Internal SIMBA constants - not exposed to users
SIMBA_DEMO_INPUT_MAX_CHARS: int = 256
